@@ -26,6 +26,7 @@ import (
 	"net/url"
 	"os"
 	"reflect"
+	"runtime"
 	"sort"
 	"strings"
 	"testing"
@@ -34,13 +35,14 @@ import (
 	"github.com/go-kit/kit/log"
 	"github.com/gogo/protobuf/proto"
 	"github.com/golang/snappy"
+	"github.com/pkg/errors"
 	"github.com/prometheus/client_golang/prometheus"
 	config_util "github.com/prometheus/common/config"
 	"github.com/prometheus/common/model"
 	"github.com/prometheus/common/promlog"
 	"github.com/prometheus/common/route"
+	"github.com/stretchr/testify/require"
 
-	"github.com/pkg/errors"
 	"github.com/prometheus/prometheus/config"
 	"github.com/prometheus/prometheus/pkg/gate"
 	"github.com/prometheus/prometheus/pkg/labels"
@@ -55,7 +57,6 @@ import (
 	"github.com/prometheus/prometheus/storage/remote"
 	"github.com/prometheus/prometheus/tsdb"
 	"github.com/prometheus/prometheus/util/teststorage"
-	"github.com/prometheus/prometheus/util/testutil"
 )
 
 // testMetaStore satisfies the scrape.MetricMetadataStore interface.
@@ -302,11 +303,16 @@ func TestEndpoints(t *testing.T) {
 			test_metric1{foo="bar"} 0+100x100
 			test_metric1{foo="boo"} 1+0x100
 			test_metric2{foo="boo"} 1+0x100
+			test_metric3{foo="bar", dup="1"} 1+0x100
+			test_metric3{foo="boo", dup="1"} 1+0x100
+			test_metric4{foo="bar", dup="1"} 1+0x100
+			test_metric4{foo="boo", dup="1"} 1+0x100
+			test_metric4{foo="boo"} 1+0x100
 	`)
-	testutil.Ok(t, err)
+	require.NoError(t, err)
 	defer suite.Close()
 
-	testutil.Ok(t, suite.Run())
+	require.NoError(t, suite.Run())
 
 	now := time.Now()
 
@@ -343,13 +349,13 @@ func TestEndpoints(t *testing.T) {
 		defer server.Close()
 
 		u, err := url.Parse(server.URL)
-		testutil.Ok(t, err)
+		require.NoError(t, err)
 
 		al := promlog.AllowedLevel{}
-		testutil.Ok(t, al.Set("debug"))
+		require.NoError(t, al.Set("debug"))
 
 		af := promlog.AllowedFormat{}
-		testutil.Ok(t, af.Set("logfmt"))
+		require.NoError(t, af.Set("logfmt"))
 
 		promlogConfig := promlog.Config{
 			Level:  &al,
@@ -357,12 +363,12 @@ func TestEndpoints(t *testing.T) {
 		}
 
 		dbDir, err := ioutil.TempDir("", "tsdb-api-ready")
-		testutil.Ok(t, err)
+		require.NoError(t, err)
 		defer os.RemoveAll(dbDir)
 
 		remote := remote.NewStorage(promlog.New(&promlogConfig), prometheus.DefaultRegisterer, func() (int64, error) {
 			return 0, nil
-		}, dbDir, 1*time.Second)
+		}, dbDir, 1*time.Second, nil)
 
 		err = remote.ApplyConfig(&config.Config{
 			RemoteReadConfigs: []*config.RemoteReadConfig{
@@ -373,7 +379,7 @@ func TestEndpoints(t *testing.T) {
 				},
 			},
 		})
-		testutil.Ok(t, err)
+		require.NoError(t, err)
 
 		var algr rulesRetrieverMock
 		algr.testing = t
@@ -411,9 +417,9 @@ func TestLabelNames(t *testing.T) {
 			test_metric2{foo="boo"} 1+0x100
 			test_metric2{foo="boo", xyz="qwerty"} 1+0x100
 	`)
-	testutil.Ok(t, err)
+	require.NoError(t, err)
 	defer suite.Close()
-	testutil.Ok(t, suite.Run())
+	require.NoError(t, suite.Run())
 
 	api := &API{
 		Queryable: suite.Storage(),
@@ -429,7 +435,7 @@ func TestLabelNames(t *testing.T) {
 	for _, method := range []string{http.MethodGet, http.MethodPost} {
 		ctx := context.Background()
 		req, err := request(method)
-		testutil.Ok(t, err)
+		require.NoError(t, err)
 		res := api.labelNames(req.WithContext(ctx))
 		assertAPIError(t, res.err, "")
 		assertAPIResponse(t, res.data, []string{"__name__", "baz", "foo", "foo1", "foo2", "xyz"})
@@ -517,12 +523,8 @@ func setupRemote(s storage.Storage) *httptest.Server {
 			}
 			defer querier.Close()
 
-			set, _, err := querier.Select(false, hints, matchers...)
-			if err != nil {
-				http.Error(w, err.Error(), http.StatusInternalServerError)
-				return
-			}
-			resp.Results[i], err = remote.ToQueryResult(set, 1e6)
+			set := querier.Select(false, hints, matchers...)
+			resp.Results[i], _, err = remote.ToQueryResult(set, 1e6)
 			if err != nil {
 				http.Error(w, err.Error(), http.StatusInternalServerError)
 				return
@@ -724,6 +726,13 @@ func testEndpoints(t *testing.T, api *API, tr *testTargetRetriever, testLabelAPI
 				labels.FromStrings("__name__", "test_metric2", "foo", "boo"),
 			},
 		},
+		{
+			endpoint: api.series,
+			query: url.Values{
+				"match[]": []string{`{foo=""}`},
+			},
+			errType: errorBadData,
+		},
 		{
 			endpoint: api.series,
 			query: url.Values{
@@ -742,6 +751,18 @@ func testEndpoints(t *testing.T, api *API, tr *testTargetRetriever, testLabelAPI
 				labels.FromStrings("__name__", "test_metric1", "foo", "boo"),
 			},
 		},
+		// Try to overlap the selected series set as much as possible to test the result de-duplication works well.
+		{
+			endpoint: api.series,
+			query: url.Values{
+				"match[]": []string{`test_metric4{foo=~".+o$"}`, `test_metric4{dup=~"^1"}`},
+			},
+			response: []labels.Labels{
+				labels.FromStrings("__name__", "test_metric4", "dup", "1", "foo", "bar"),
+				labels.FromStrings("__name__", "test_metric4", "dup", "1", "foo", "boo"),
+				labels.FromStrings("__name__", "test_metric4", "foo", "boo"),
+			},
+		},
 		{
 			endpoint: api.series,
 			query: url.Values{
@@ -1197,7 +1218,7 @@ func testEndpoints(t *testing.T, api *API, tr *testTargetRetriever, testLabelAPI
 				},
 			},
 			response: map[string][]metadata{
-				"go_threads": []metadata{
+				"go_threads": {
 					{textparse.MetricTypeGauge, "Number of OS threads created", ""},
 					{textparse.MetricTypeGauge, "Number of OS threads that were created.", ""},
 				},
@@ -1283,7 +1304,7 @@ func testEndpoints(t *testing.T, api *API, tr *testTargetRetriever, testLabelAPI
 				},
 			},
 			response: map[string][]metadata{
-				"go_threads": []metadata{
+				"go_threads": {
 					{textparse.MetricTypeGauge, "Number of OS threads created", ""},
 					{textparse.MetricTypeGauge, "Number of OS threads that were created.", ""},
 				},
@@ -1454,6 +1475,8 @@ func testEndpoints(t *testing.T, api *API, tr *testTargetRetriever, testLabelAPI
 				response: []string{
 					"test_metric1",
 					"test_metric2",
+					"test_metric3",
+					"test_metric4",
 				},
 			},
 			{
@@ -1598,11 +1621,102 @@ func testEndpoints(t *testing.T, api *API, tr *testTargetRetriever, testLabelAPI
 					"boo",
 				},
 			},
-
+			// Label values with bad matchers.
+			{
+				endpoint: api.labelValues,
+				params: map[string]string{
+					"name": "foo",
+				},
+				query: url.Values{
+					"match[]": []string{`{foo=""`, `test_metric2`},
+				},
+				errType: errorBadData,
+			},
+			// Label values with empty matchers.
+			{
+				endpoint: api.labelValues,
+				params: map[string]string{
+					"name": "foo",
+				},
+				query: url.Values{
+					"match[]": []string{`{foo=""}`},
+				},
+				errType: errorBadData,
+			},
+			// Label values with matcher.
+			{
+				endpoint: api.labelValues,
+				params: map[string]string{
+					"name": "foo",
+				},
+				query: url.Values{
+					"match[]": []string{`test_metric2`},
+				},
+				response: []string{
+					"boo",
+				},
+			},
+			// Label values with matcher.
+			{
+				endpoint: api.labelValues,
+				params: map[string]string{
+					"name": "foo",
+				},
+				query: url.Values{
+					"match[]": []string{`test_metric1`},
+				},
+				response: []string{
+					"bar",
+					"boo",
+				},
+			},
+			// Label values with matcher using label filter.
+			{
+				endpoint: api.labelValues,
+				params: map[string]string{
+					"name": "foo",
+				},
+				query: url.Values{
+					"match[]": []string{`test_metric1{foo="bar"}`},
+				},
+				response: []string{
+					"bar",
+				},
+			},
+			// Label values with matcher and time range.
+			{
+				endpoint: api.labelValues,
+				params: map[string]string{
+					"name": "foo",
+				},
+				query: url.Values{
+					"match[]": []string{`test_metric1`},
+					"start":   []string{"1"},
+					"end":     []string{"100000000"},
+				},
+				response: []string{
+					"bar",
+					"boo",
+				},
+			},
+			// Try to overlap the selected series set as much as possible to test that the value de-duplication works.
+			{
+				endpoint: api.labelValues,
+				params: map[string]string{
+					"name": "foo",
+				},
+				query: url.Values{
+					"match[]": []string{`test_metric4{dup=~"^1"}`, `test_metric4{foo=~".+o$"}`},
+				},
+				response: []string{
+					"bar",
+					"boo",
+				},
+			},
 			// Label names.
 			{
 				endpoint: api.labelNames,
-				response: []string{"__name__", "foo"},
+				response: []string{"__name__", "dup", "foo"},
 			},
 			// Start and end before Label names starts.
 			{
@@ -1620,7 +1734,7 @@ func testEndpoints(t *testing.T, api *API, tr *testTargetRetriever, testLabelAPI
 					"start": []string{"1"},
 					"end":   []string{"100"},
 				},
-				response: []string{"__name__", "foo"},
+				response: []string{"__name__", "dup", "foo"},
 			},
 			// Start before Label names, end within Label names.
 			{
@@ -1629,7 +1743,7 @@ func testEndpoints(t *testing.T, api *API, tr *testTargetRetriever, testLabelAPI
 					"start": []string{"-1"},
 					"end":   []string{"10"},
 				},
-				response: []string{"__name__", "foo"},
+				response: []string{"__name__", "dup", "foo"},
 			},
 
 			// Start before Label names starts, end after Label names ends.
@@ -1639,7 +1753,7 @@ func testEndpoints(t *testing.T, api *API, tr *testTargetRetriever, testLabelAPI
 					"start": []string{"-1"},
 					"end":   []string{"100000"},
 				},
-				response: []string{"__name__", "foo"},
+				response: []string{"__name__", "dup", "foo"},
 			},
 			// Start with bad data for Label names, end within Label names.
 			{
@@ -1657,7 +1771,7 @@ func testEndpoints(t *testing.T, api *API, tr *testTargetRetriever, testLabelAPI
 					"start": []string{"1"},
 					"end":   []string{"1000000006"},
 				},
-				response: []string{"__name__", "foo"},
+				response: []string{"__name__", "dup", "foo"},
 			},
 			// Start and end after Label names ends.
 			{
@@ -1674,7 +1788,7 @@ func testEndpoints(t *testing.T, api *API, tr *testTargetRetriever, testLabelAPI
 				query: url.Values{
 					"start": []string{"4"},
 				},
-				response: []string{"__name__", "foo"},
+				response: []string{"__name__", "dup", "foo"},
 			},
 			// Only provide End within Label names, don't provide a start time.
 			{
@@ -1682,6 +1796,60 @@ func testEndpoints(t *testing.T, api *API, tr *testTargetRetriever, testLabelAPI
 				query: url.Values{
 					"end": []string{"20"},
 				},
+				response: []string{"__name__", "dup", "foo"},
+			},
+			// Label names with bad matchers.
+			{
+				endpoint: api.labelNames,
+				query: url.Values{
+					"match[]": []string{`{foo=""`, `test_metric2`},
+				},
+				errType: errorBadData,
+			},
+			// Label values with empty matchers.
+			{
+				endpoint: api.labelNames,
+				params: map[string]string{
+					"name": "foo",
+				},
+				query: url.Values{
+					"match[]": []string{`{foo=""}`},
+				},
+				errType: errorBadData,
+			},
+			// Label names with matcher.
+			{
+				endpoint: api.labelNames,
+				query: url.Values{
+					"match[]": []string{`test_metric2`},
+				},
+				response: []string{"__name__", "foo"},
+			},
+			// Label names with matcher.
+			{
+				endpoint: api.labelNames,
+				query: url.Values{
+					"match[]": []string{`test_metric3`},
+				},
+				response: []string{"__name__", "dup", "foo"},
+			},
+			// Label names with matcher using label filter.
+			// There is no matching series.
+			{
+				endpoint: api.labelNames,
+				query: url.Values{
+					"match[]": []string{`test_metric1{foo="test"}`},
+				},
+				response: []string{},
+			},
+			// Label names with matcher and time range.
+			{
+				endpoint: api.labelNames,
+				query: url.Values{
+					"match[]": []string{`test_metric2`},
+					"start":   []string{"1"},
+					"end":     []string{"100000000"},
+				},
 				response: []string{"__name__", "foo"},
 			},
 		}...)
@@ -1708,40 +1876,48 @@ func testEndpoints(t *testing.T, api *API, tr *testTargetRetriever, testLabelAPI
 	}
 
 	for i, test := range tests {
-		for _, method := range methods(test.endpoint) {
-			// Build a context with the correct request params.
-			ctx := context.Background()
-			for p, v := range test.params {
-				ctx = route.WithParam(ctx, p, v)
-			}
-			t.Logf("run %d\t%s\t%q", i, method, test.query.Encode())
+		t.Run(fmt.Sprintf("run %d %s %q", i, describeAPIFunc(test.endpoint), test.query.Encode()), func(t *testing.T) {
+			for _, method := range methods(test.endpoint) {
+				t.Run(method, func(t *testing.T) {
+					// Build a context with the correct request params.
+					ctx := context.Background()
+					for p, v := range test.params {
+						ctx = route.WithParam(ctx, p, v)
+					}
 
-			req, err := request(method, test.query)
-			if err != nil {
-				t.Fatal(err)
-			}
+					req, err := request(method, test.query)
+					if err != nil {
+						t.Fatal(err)
+					}
 
-			tr.ResetMetadataStore()
-			for _, tm := range test.metadata {
-				tr.SetMetadataStoreForTargets(tm.identifier, &testMetaStore{Metadata: tm.metadata})
-			}
+					tr.ResetMetadataStore()
+					for _, tm := range test.metadata {
+						tr.SetMetadataStoreForTargets(tm.identifier, &testMetaStore{Metadata: tm.metadata})
+					}
 
-			res := test.endpoint(req.WithContext(ctx))
-			assertAPIError(t, res.err, test.errType)
+					res := test.endpoint(req.WithContext(ctx))
+					assertAPIError(t, res.err, test.errType)
 
-			if test.sorter != nil {
-				test.sorter(res.data)
-			}
+					if test.sorter != nil {
+						test.sorter(res.data)
+					}
 
-			if test.responseLen != 0 {
-				assertAPIResponseLength(t, res.data, test.responseLen)
-			} else {
-				assertAPIResponse(t, res.data, test.response)
+					if test.responseLen != 0 {
+						assertAPIResponseLength(t, res.data, test.responseLen)
+					} else {
+						assertAPIResponse(t, res.data, test.response)
+					}
+				})
 			}
-		}
+		})
 	}
 }
 
+func describeAPIFunc(f apiFunc) string {
+	name := runtime.FuncForPC(reflect.ValueOf(f).Pointer()).Name()
+	return strings.Split(name[strings.LastIndex(name, ".")+1:], "-")[0]
+}
+
 func assertAPIError(t *testing.T, got *apiError, exp errorType) {
 	t.Helper()
 
@@ -1762,23 +1938,7 @@ func assertAPIError(t *testing.T, got *apiError, exp errorType) {
 func assertAPIResponse(t *testing.T, got interface{}, exp interface{}) {
 	t.Helper()
 
-	if !reflect.DeepEqual(exp, got) {
-		respJSON, err := json.Marshal(got)
-		if err != nil {
-			t.Fatalf("failed to marshal response as JSON: %v", err.Error())
-		}
-
-		expectedRespJSON, err := json.Marshal(exp)
-		if err != nil {
-			t.Fatalf("failed to marshal expected response as JSON: %v", err.Error())
-		}
-
-		t.Fatalf(
-			"Response does not match, expected:\n%+v\ngot:\n%+v",
-			string(expectedRespJSON),
-			string(respJSON),
-		)
-	}
+	require.Equal(t, exp, got)
 }
 
 func assertAPIResponseLength(t *testing.T, got interface{}, expLen int) {
@@ -1799,12 +1959,12 @@ func TestSampledReadEndpoint(t *testing.T) {
 		load 1m
 			test_metric1{foo="bar",baz="qux"} 1
 	`)
-	testutil.Ok(t, err)
+	require.NoError(t, err)
 
 	defer suite.Close()
 
 	err = suite.Run()
-	testutil.Ok(t, err)
+	require.NoError(t, err)
 
 	api := &API{
 		Queryable:   suite.Storage(),
@@ -1827,21 +1987,21 @@ func TestSampledReadEndpoint(t *testing.T) {
 
 	// Encode the request.
 	matcher1, err := labels.NewMatcher(labels.MatchEqual, "__name__", "test_metric1")
-	testutil.Ok(t, err)
+	require.NoError(t, err)
 
 	matcher2, err := labels.NewMatcher(labels.MatchEqual, "d", "e")
-	testutil.Ok(t, err)
+	require.NoError(t, err)
 
 	query, err := remote.ToQuery(0, 1, []*labels.Matcher{matcher1, matcher2}, &storage.SelectHints{Step: 0, Func: "avg"})
-	testutil.Ok(t, err)
+	require.NoError(t, err)
 
 	req := &prompb.ReadRequest{Queries: []*prompb.Query{query}}
 	data, err := proto.Marshal(req)
-	testutil.Ok(t, err)
+	require.NoError(t, err)
 
 	compressed := snappy.Encode(nil, data)
 	request, err := http.NewRequest("POST", "", bytes.NewBuffer(compressed))
-	testutil.Ok(t, err)
+	require.NoError(t, err)
 
 	recorder := httptest.NewRecorder()
 	api.remoteRead(recorder, request)
@@ -1850,25 +2010,25 @@ func TestSampledReadEndpoint(t *testing.T) {
 		t.Fatal(recorder.Code)
 	}
 
-	testutil.Equals(t, "application/x-protobuf", recorder.Result().Header.Get("Content-Type"))
-	testutil.Equals(t, "snappy", recorder.Result().Header.Get("Content-Encoding"))
+	require.Equal(t, "application/x-protobuf", recorder.Result().Header.Get("Content-Type"))
+	require.Equal(t, "snappy", recorder.Result().Header.Get("Content-Encoding"))
 
 	// Decode the response.
 	compressed, err = ioutil.ReadAll(recorder.Result().Body)
-	testutil.Ok(t, err)
+	require.NoError(t, err)
 
 	uncompressed, err := snappy.Decode(nil, compressed)
-	testutil.Ok(t, err)
+	require.NoError(t, err)
 
 	var resp prompb.ReadResponse
 	err = proto.Unmarshal(uncompressed, &resp)
-	testutil.Ok(t, err)
+	require.NoError(t, err)
 
 	if len(resp.Results) != 1 {
 		t.Fatalf("Expected 1 result, got %d", len(resp.Results))
 	}
 
-	testutil.Equals(t, &prompb.QueryResult{
+	require.Equal(t, &prompb.QueryResult{
 		Timeseries: []*prompb.TimeSeries{
 			{
 				Labels: []prompb.Label{
@@ -1894,11 +2054,11 @@ func TestStreamReadEndpoint(t *testing.T) {
             test_metric1{foo="bar2",baz="qux"} 0+100x120
             test_metric1{foo="bar3",baz="qux"} 0+100x240
 	`)
-	testutil.Ok(t, err)
+	require.NoError(t, err)
 
 	defer suite.Close()
 
-	testutil.Ok(t, suite.Run())
+	require.NoError(t, suite.Run())
 
 	api := &API{
 		Queryable:   suite.Storage(),
@@ -1923,13 +2083,13 @@ func TestStreamReadEndpoint(t *testing.T) {
 
 	// Encode the request.
 	matcher1, err := labels.NewMatcher(labels.MatchEqual, "__name__", "test_metric1")
-	testutil.Ok(t, err)
+	require.NoError(t, err)
 
 	matcher2, err := labels.NewMatcher(labels.MatchEqual, "d", "e")
-	testutil.Ok(t, err)
+	require.NoError(t, err)
 
 	matcher3, err := labels.NewMatcher(labels.MatchEqual, "foo", "bar1")
-	testutil.Ok(t, err)
+	require.NoError(t, err)
 
 	query1, err := remote.ToQuery(0, 14400001, []*labels.Matcher{matcher1, matcher2}, &storage.SelectHints{
 		Step:  1,
@@ -1937,7 +2097,7 @@ func TestStreamReadEndpoint(t *testing.T) {
 		Start: 0,
 		End:   14400001,
 	})
-	testutil.Ok(t, err)
+	require.NoError(t, err)
 
 	query2, err := remote.ToQuery(0, 14400001, []*labels.Matcher{matcher1, matcher3}, &storage.SelectHints{
 		Step:  1,
@@ -1945,18 +2105,18 @@ func TestStreamReadEndpoint(t *testing.T) {
 		Start: 0,
 		End:   14400001,
 	})
-	testutil.Ok(t, err)
+	require.NoError(t, err)
 
 	req := &prompb.ReadRequest{
 		Queries:               []*prompb.Query{query1, query2},
 		AcceptedResponseTypes: []prompb.ReadRequest_ResponseType{prompb.ReadRequest_STREAMED_XOR_CHUNKS},
 	}
 	data, err := proto.Marshal(req)
-	testutil.Ok(t, err)
+	require.NoError(t, err)
 
 	compressed := snappy.Encode(nil, data)
 	request, err := http.NewRequest("POST", "", bytes.NewBuffer(compressed))
-	testutil.Ok(t, err)
+	require.NoError(t, err)
 
 	recorder := httptest.NewRecorder()
 	api.remoteRead(recorder, request)
@@ -1965,8 +2125,8 @@ func TestStreamReadEndpoint(t *testing.T) {
 		t.Fatal(recorder.Code)
 	}
 
-	testutil.Equals(t, "application/x-streamed-protobuf; proto=prometheus.ChunkedReadResponse", recorder.Result().Header.Get("Content-Type"))
-	testutil.Equals(t, "", recorder.Result().Header.Get("Content-Encoding"))
+	require.Equal(t, "application/x-streamed-protobuf; proto=prometheus.ChunkedReadResponse", recorder.Result().Header.Get("Content-Type"))
+	require.Equal(t, "", recorder.Result().Header.Get("Content-Encoding"))
 
 	var results []*prompb.ChunkedReadResponse
 	stream := remote.NewChunkedReader(recorder.Result().Body, remote.DefaultChunkedReadLimit, nil)
@@ -1976,7 +2136,7 @@ func TestStreamReadEndpoint(t *testing.T) {
 		if err == io.EOF {
 			break
 		}
-		testutil.Ok(t, err)
+		require.NoError(t, err)
 		results = append(results, res)
 	}
 
@@ -1984,7 +2144,7 @@ func TestStreamReadEndpoint(t *testing.T) {
 		t.Fatalf("Expected 5 result, got %d", len(results))
 	}
 
-	testutil.Equals(t, []*prompb.ChunkedReadResponse{
+	require.Equal(t, []*prompb.ChunkedReadResponse{
 		{
 			ChunkedSeries: []*prompb.ChunkedSeries{
 				{
@@ -2120,7 +2280,9 @@ func (f *fakeDB) Stats(statsByLabelName string) (_ *tsdb.Stats, retErr error) {
 			retErr = err
 		}
 	}()
-	h, _ := tsdb.NewHead(nil, nil, nil, 1000, "", nil, tsdb.DefaultStripeSize, nil)
+	opts := tsdb.DefaultHeadOptions()
+	opts.ChunkRange = 1000
+	h, _ := tsdb.NewHead(nil, nil, nil, opts)
 	return h.Stats(statsByLabelName), nil
 }
 
@@ -2288,7 +2450,7 @@ func TestAdminEndpoints(t *testing.T) {
 		tc := tc
 		t.Run("", func(t *testing.T) {
 			dir, _ := ioutil.TempDir("", "fakeDB")
-			defer testutil.Ok(t, os.RemoveAll(dir))
+			defer func() { require.NoError(t, os.RemoveAll(dir)) }()
 
 			api := &API{
 				db:          tc.db,
@@ -2299,7 +2461,7 @@ func TestAdminEndpoints(t *testing.T) {
 
 			endpoint := tc.endpoint(api)
 			req, err := http.NewRequest(tc.method, fmt.Sprintf("?%s", tc.values.Encode()), nil)
-			testutil.Ok(t, err)
+			require.NoError(t, err)
 
 			res := setUnavailStatusOnTSDBNotReady(endpoint(req))
 			assertAPIError(t, res.err, tc.errType)
@@ -2340,9 +2502,7 @@ func TestRespondSuccess(t *testing.T) {
 		Status: statusSuccess,
 		Data:   "test",
 	}
-	if !reflect.DeepEqual(&res, exp) {
-		t.Fatalf("Expected response \n%v\n but got \n%v\n", res, exp)
-	}
+	require.Equal(t, exp, &res)
 }
 
 func TestRespondError(t *testing.T) {
@@ -2380,9 +2540,7 @@ func TestRespondError(t *testing.T) {
 		ErrorType: errorTimeout,
 		Error:     "message",
 	}
-	if !reflect.DeepEqual(&res, exp) {
-		t.Fatalf("Expected response \n%v\n but got \n%v\n", res, exp)
-	}
+	require.Equal(t, exp, &res)
 }
 
 func TestParseTimeParam(t *testing.T) {
@@ -2392,7 +2550,7 @@ func TestParseTimeParam(t *testing.T) {
 	}
 
 	ts, err := parseTime("1582468023986")
-	testutil.Ok(t, err)
+	require.NoError(t, err)
 
 	var tests = []struct {
 		paramName    string
@@ -2434,15 +2592,15 @@ func TestParseTimeParam(t *testing.T) {
 
 	for _, test := range tests {
 		req, err := http.NewRequest("GET", "localhost:42/foo?"+test.paramName+"="+test.paramValue, nil)
-		testutil.Ok(t, err)
+		require.NoError(t, err)
 
 		result := test.result
 		asTime, err := parseTimeParam(req, test.paramName, test.defaultValue)
 
 		if err != nil {
-			testutil.ErrorEqual(t, result.asError(), err)
+			require.EqualError(t, err, result.asError().Error())
 		} else {
-			testutil.Assert(t, asTime.Equal(result.asTime), "time as return value: %s not parsed correctly. Expected %s. Actual %s", test.paramValue, result.asTime, asTime)
+			require.True(t, asTime.Equal(result.asTime), "time as return value: %s not parsed correctly. Expected %s. Actual %s", test.paramValue, result.asTime, asTime)
 		}
 	}
 }
@@ -2711,7 +2869,7 @@ func TestTSDBStatus(t *testing.T) {
 	} {
 		tc := tc
 		t.Run(fmt.Sprintf("%d", i), func(t *testing.T) {
-			api := &API{db: tc.db}
+			api := &API{db: tc.db, gatherer: prometheus.DefaultGatherer}
 			endpoint := tc.endpoint(api)
 			req, err := http.NewRequest(tc.method, fmt.Sprintf("?%s", tc.values.Encode()), nil)
 			if err != nil {
@@ -2723,6 +2881,42 @@ func TestTSDBStatus(t *testing.T) {
 	}
 }
 
+func TestReturnAPIError(t *testing.T) {
+	cases := []struct {
+		err      error
+		expected errorType
+	}{
+		{
+			err:      promql.ErrStorage{Err: errors.New("storage error")},
+			expected: errorInternal,
+		}, {
+			err:      errors.Wrap(promql.ErrStorage{Err: errors.New("storage error")}, "wrapped"),
+			expected: errorInternal,
+		}, {
+			err:      promql.ErrQueryTimeout("timeout error"),
+			expected: errorTimeout,
+		}, {
+			err:      errors.Wrap(promql.ErrQueryTimeout("timeout error"), "wrapped"),
+			expected: errorTimeout,
+		}, {
+			err:      promql.ErrQueryCanceled("canceled error"),
+			expected: errorCanceled,
+		}, {
+			err:      errors.Wrap(promql.ErrQueryCanceled("canceled error"), "wrapped"),
+			expected: errorCanceled,
+		}, {
+			err:      errors.New("exec error"),
+			expected: errorExec,
+		},
+	}
+
+	for _, c := range cases {
+		actual := returnAPIError(c.err)
+		require.Error(t, actual)
+		require.Equal(t, c.expected, actual.typ)
+	}
+}
+
 // This is a global to avoid the benchmark being optimized away.
 var testResponseWriter = httptest.ResponseRecorder{}
 
@@ -2747,3 +2941,111 @@ func BenchmarkRespond(b *testing.B) {
 		api.respond(&testResponseWriter, response, nil)
 	}
 }
+
+func TestGetGlobalURL(t *testing.T) {
+	mustParseURL := func(t *testing.T, u string) *url.URL {
+		parsed, err := url.Parse(u)
+		require.NoError(t, err)
+		return parsed
+	}
+
+	testcases := []struct {
+		input    *url.URL
+		opts     GlobalURLOptions
+		expected *url.URL
+		errorful bool
+	}{
+		{
+			mustParseURL(t, "http://127.0.0.1:9090"),
+			GlobalURLOptions{
+				ListenAddress: "127.0.0.1:9090",
+				Host:          "127.0.0.1:9090",
+				Scheme:        "http",
+			},
+			mustParseURL(t, "http://127.0.0.1:9090"),
+			false,
+		},
+		{
+			mustParseURL(t, "http://127.0.0.1:9090"),
+			GlobalURLOptions{
+				ListenAddress: "127.0.0.1:9090",
+				Host:          "prometheus.io",
+				Scheme:        "https",
+			},
+			mustParseURL(t, "https://prometheus.io"),
+			false,
+		},
+		{
+			mustParseURL(t, "http://exemple.com"),
+			GlobalURLOptions{
+				ListenAddress: "127.0.0.1:9090",
+				Host:          "prometheus.io",
+				Scheme:        "https",
+			},
+			mustParseURL(t, "http://exemple.com"),
+			false,
+		},
+		{
+			mustParseURL(t, "http://localhost:8080"),
+			GlobalURLOptions{
+				ListenAddress: "127.0.0.1:9090",
+				Host:          "prometheus.io",
+				Scheme:        "https",
+			},
+			mustParseURL(t, "http://prometheus.io:8080"),
+			false,
+		},
+		{
+			mustParseURL(t, "http://[::1]:8080"),
+			GlobalURLOptions{
+				ListenAddress: "127.0.0.1:9090",
+				Host:          "prometheus.io",
+				Scheme:        "https",
+			},
+			mustParseURL(t, "http://prometheus.io:8080"),
+			false,
+		},
+		{
+			mustParseURL(t, "http://localhost"),
+			GlobalURLOptions{
+				ListenAddress: "127.0.0.1:9090",
+				Host:          "prometheus.io",
+				Scheme:        "https",
+			},
+			mustParseURL(t, "http://prometheus.io"),
+			false,
+		},
+		{
+			mustParseURL(t, "http://localhost:9091"),
+			GlobalURLOptions{
+				ListenAddress: "[::1]:9090",
+				Host:          "[::1]",
+				Scheme:        "https",
+			},
+			mustParseURL(t, "http://[::1]:9091"),
+			false,
+		},
+		{
+			mustParseURL(t, "http://localhost:9091"),
+			GlobalURLOptions{
+				ListenAddress: "[::1]:9090",
+				Host:          "[::1]:9090",
+				Scheme:        "https",
+			},
+			mustParseURL(t, "http://[::1]:9091"),
+			false,
+		},
+	}
+
+	for i, tc := range testcases {
+		t.Run(fmt.Sprintf("Test %d", i), func(t *testing.T) {
+			output, err := getGlobalURL(tc.input, tc.opts)
+			if tc.errorful {
+				require.Error(t, err)
+				return
+			}
+			require.NoError(t, err)
+			require.Equal(t, tc.expected, output)
+		})
+	}
+}
